// Generated by ResEdit 1.4.13
// Copyright (C) 2006-2010
// http://www.resedit.net

#include <windows.h>
#include <commctrl.h>
#include <richedit.h>
#include "resource.h"




//
// Icon resources
//
IDI_MAINICON       ICON           "..\\icon\\256x256.ico"
