#ifndef ZAP_H
#define ZAP_H

void zap_instruments(void* no_confirm, void* b, void* c);
void zap_sequence(void* no_confirm, void* b, void* c);
void zap_fx(void* no_confirm, void* b, void* c);
void zap_wavetable(void* no_confirm, void* b, void* c);

#endif
