EXPORTS
KSND_LoadSong
KSND_LoadSongFromMemory
KSND_FreeSong
KSND_GetSongLength
KSND_CreatePlayer
KSND_CreatePlayerUnregistered
KSND_SetPlayerQuality
KSND_FreePlayer
KSND_PlaySong
KSND_FillBuffer
KSND_Stop
KSND_Pause
KSND_GetPlayPosition
KSND_SetVolume
KSND_GetVUMeters
KSND_GetSongInfo
KSND_SetLooping
KSND_GetPlayTime
